(*
** This file has been generated, if you wish to
** modify it in a permanent way, please refer
** to the script file : gen/generator_caml.rb
*)

open Api


(* utilitaires *)

let nonzero_int_sign n = n / abs n

let flush_outputs () =
  flush stderr; flush stdout (* Pour que vos sorties s'affichent *)

let foldl = List.fold_left
let foldr = List.fold_right
let foldl1 f (x::xs) = foldl f x xs

let nonempty_list_to_array ((x::_) as l) =
  let t = Array.make (List.length l) x in
  let rec loop i = function
    | [] -> ()
    | y::ys -> t.(i) <- y; loop (i+1) ys in
  loop 0 l;
  t

let in_place_shuffle t =
  let n = Array.length t in
  let swap j k = let tmp = t.(j) in
                 t.(j) <- t.(k); t.(k) <- tmp in
  for i = 0 to n-1 do
    swap i (i + Random.int (n-i))
  done;
  t

let shuffle_list l =
  if l = [] then [] else
    Array.to_list (in_place_shuffle (nonempty_list_to_array l))

let rec take n l = match (n,l) with
  | 0, _ -> []
  | n, x::xs -> x :: take (n-1) xs
  | _, [] -> []

let error_message = function
  | Ok -> "L'action a été exécutée avec succès"
  | Or_insuffisant -> "Vous ne possédez pas assez d'or pour cette action"
  | Ile_invalide -> "La position spécifiée n'est pas une île"
  | Position_invalide -> "La position spécifiée n'est pas valide"
  | Trop_loin -> "La destination est trop éloignée"
  | Ile_colonisee -> "L'île est déjà colonisée"
  | Ile_ennemie -> "L'île ne vous appartient pas"
  | Bateau_ennemi -> "Le bateau ne vous appartient pas"
  | Id_invalide -> "L'ID spécifiée n'est pas valide"
  | Non_deplacable -> "Le bateau n'est pas déplaçable"
  | Aucune_caravelle -> "Il n'y a aucune caravelle susceptible de coloniser l'île"
  | Limite_atteinte -> "La limite de bateaux est atteinte"

let log_error label err =
  if err <> Ok then (
    print_string (label ^ ": ");
    print_endline (error_message err)
  )


(* globales et fcts auxiliaires simples *)

let ile_depart = ref (0,0)
let caravelle_depart = ref (-1)
let mon_id = ref 0 and adversaire_id = ref 0


let portee = function
  | Bateau_caravelle -> caravelle_deplacement
  | Bateau_galion -> galion_deplacement
  | Bateau_erreur -> assert false

let rapprocher_max bat_id ((x,y) as dest) =
  let bat_data = info_bateau bat_id in
  let range = portee bat_data.btype in
  let (xb, yb) as bpos = bat_data.pos in
  if distance bpos dest <= range
  then deplacer bat_id dest
  else let delta_x = x - xb and delta_y = y - yb in
       let target =
         if abs delta_x <= range
         then (x, (nonzero_int_sign delta_y)*(range - abs delta_x) + yb)
         else ((nonzero_int_sign delta_x)*range + xb, yb) in
       deplacer bat_id target

(*
let enemy_islands () =
  List.filter (fun ile -> ile_info_joueur ile = adversaire_id)
              (Array.to_list (liste_iles ()))

*)


let for_pos_in_terrain f = 
  for x = 0 to taille_terrain - 1 do
    for y = 0 to taille_terrain - 1 do
      f x y
    done
  done

let norm_1_ball (x,y) =
  let r = portee Bateau_galion and l = ref [] in
  for i = max (x-r) 0 to min (x+r) (taille_terrain-1) do
    for j = max (y-r) 0 to min (y+r) (taille_terrain-1) do
      if distance (x,y) (i,j) <= r
      then l := (i,j) :: !l
    done
  done;
  shuffle_list !l

let for_pos_in_range center f =
  List.iter (fun (i,j) -> f i j) (norm_1_ball center)

let force_frappe player (x,y) =
  let acc = ref [] in
  for_pos_in_range (x,y) (fun i j ->
    let is_threat boat = boat.joueur = player
                       && boat.btype = Bateau_galion
                         && ( (i,j) = (x,y) || boat.deplacable ) in
    acc := List.filter is_threat
                       (Array.to_list (liste_bateaux_position (i,j)))
                       @ !acc
  );
  !acc



(* opening strategy choice predicate *)

type opening = Rush | LowEcon | NormalEcon

let opening_choice = match Array.length (liste_iles ()) with
  | 2 -> Rush
  | n when n < 10 -> LowEcon
  | _ -> NormalEcon


(*
** Fonction appelée au début de la partie
*)
let partie_init () =

  Random.self_init ();

  ile_depart := (mes_iles ()).(0);
  mon_id := mon_joueur ();
  adversaire_id := adversaire ();

  flush_outputs ()


(*
** Fonction appelée à chaque tour
*)
let jouer_tour () =

  (* Opening strategies *)
  (match tour_actuel () with
  | 1 ->  log_error "fst crv" (construire Bateau_caravelle !ile_depart);
          caravelle_depart := id_dernier_bateau_construit ()
  | _ -> ()
  );

  if bateau_existe !caravelle_depart then (
    
  (*** colonisation ***)
  (* 1 caravelle, priorité aux îles non volcaniques *)

    let caravelle_target =
      let is_free ile = info_ile_joueur ile = -1 in
      let free_islands = List.filter is_free (Array.to_list (liste_iles ())) in
      let pos_crv = (info_bateau !caravelle_depart).pos in
      let better_target a b =
        match (info_terrain a, info_terrain b) with
        | Terrain_ile, Terrain_volcan -> a
        | Terrain_volcan, Terrain_ile -> b
        | _,_ -> if distance a pos_crv < distance b pos_crv
                 then a else b
      in
      if free_islands = [] then None
      else Some (foldl1 better_target free_islands)
    in

    match caravelle_target with
    | None -> ()
    | Some target -> (log_error "2" (rapprocher_max !caravelle_depart target);
                      if (info_bateau !caravelle_depart).pos = target
                      then log_error "3" (coloniser target))
  );
  

  (*** attaques/déplacements ***)

  (* enemy presence on the map *)
  let enemy_galleons = Array.make_matrix taille_terrain taille_terrain 0
  and enemy_caravelles = Array.make_matrix taille_terrain taille_terrain 0 in
  for_pos_in_terrain (fun i j ->
      Array.iter (fun b -> match b.btype with
                           | Bateau_galion ->
                             enemy_galleons.(i).(j) <-
                               enemy_galleons.(i).(j)
                               + (if b.joueur = !adversaire_id then 1 else 0)
                           | Bateau_caravelle ->
                             enemy_caravelles.(i).(j) <-
                               enemy_caravelles.(i).(j)
                               + (if b.joueur = !adversaire_id then 1 else 0)
                           | Bateau_erreur -> assert false) 
                 (liste_bateaux_position (i,j))
  );

  let num_force_frappe_adv (x,y) =
    let acc = ref 0 in
    for_pos_in_range (x,y) (fun i j ->
      acc := !acc + enemy_galleons.(i).(j)
    );
    !acc
  in

  (* offensive movement function *)

  let move_all fleet pos =
    List.iter (fun b -> log_error "move_all" (deplacer b.id pos)) fleet
  in

  let attaque_totale fleet ((x,y) as ile) =
    move_all fleet ile;
    (* remember the forces have gone down for future moves *)
    enemy_caravelles.(x).(y) <- 0;
    enemy_galleons.(x).(y) <- 0 (* même s'il y a pas d'ennemi là
                                   c'est pas grave *)
  in
  

  (* capturer îles ennemies proches *)

  let maybe_capture_position pos =
    let force = force_frappe !mon_id pos in
    if List.length force > num_force_frappe_adv pos
    then attaque_totale force pos
  in

  Array.iter (fun ile -> if info_ile_joueur ile = !adversaire_id
                         then maybe_capture_position ile)
             (liste_iles ());

  (* attaquer bateaux ennemis
     ne pas attaquer les caravelles trop bien protégées
  *)

  for_pos_in_terrain (fun i j ->
    if max enemy_galleons.(i).(j) enemy_caravelles.(i).(j) > 0
    then maybe_capture_position (i,j)
  );

  (** positionnement **)

  (* critical tiles = those reachable by the enemy *)
  let critical = Array.make_matrix taille_terrain taille_terrain false in
   for_pos_in_terrain (fun x y ->
     (try for_pos_in_range (x,y) (fun i j ->
       if enemy_galleons.(i).(j) > 0 then raise Not_found)
      with Not_found -> critical.(x).(y) <- true);
     (* HACK : make the AI aware of unguarded caravelles
               as important targets *)
     if enemy_caravelles.(x).(y) > 0 then critical.(x).(y) <- true
   );
  
(*
  for i = 1 to taille_terrain-1 do
    for j = 1 to taille_terrain-1 do
      print_string (if critical.(j).(i) then "X" else "*")
    done;
    print_newline ()
  done;
*)
  (* calculate distances to the critical area using a BFS *)
  let dist_to_critical = Array.make_matrix taille_terrain taille_terrain (-1) in
  (let bfs_queue = Queue.create () in
   for_pos_in_terrain (fun x y -> if critical.(x).(y) then (
                                    dist_to_critical.(x).(y) <- 0;
                                    Queue.push (x,y) bfs_queue )
   );
   let process_node (x,y) =
     let d = dist_to_critical.(x).(y) in
     List.iter (fun (i,j) -> if min i j >= 0 && max i j <= taille_terrain - 1
                                && dist_to_critical.(i).(j) = -1
                             then ( dist_to_critical.(i).(j) <- d+1;
                                    Queue.push (i,j) bfs_queue ) )
               [(x+1,y);(x-1,y);(x,y+1);(x,y-1)]
   in
   while not (Queue.is_empty bfs_queue) do
     process_node (Queue.pop bfs_queue)
   done
  );
(*
  for i = 1 to taille_terrain-1 do
    for j = 1 to taille_terrain-1 do
      Printf.printf "%3d" dist_to_critical.(j).(i)
    done;
    print_newline ()
  done;
*)    

  (* tile value = critical tiles in range *)
  let tile_value = Array.make_matrix taille_terrain taille_terrain 0 in
  let tile_value_list = ref [] in
  for_pos_in_terrain (fun x y ->
    for_pos_in_range (x,y) (fun i j ->
      tile_value.(x).(y) <- tile_value.(x).(y)
                          + (if critical.(i).(j) then 1 else 0)
    );
    tile_value_list := (tile_value.(x).(y), (x,y)) :: !tile_value_list
  );
  
  (* tactical movements close to the enemy *)
  let tactical_positions =
    let compare_decreasing a b = compare b a in
    let sorted_by_value = List.map snd (List.sort compare_decreasing
                                                  !tile_value_list) in
    let is_interesting_position (i,j) =
      (force_frappe !mon_id (i,j) <> []) && critical.(i).(j) in
    shuffle_list (take 8 (List.filter is_interesting_position sorted_by_value))
  in
  List.iter maybe_capture_position tactical_positions;

  (* guiding the incoming stream of backup units *)
  
  for_pos_in_terrain (fun x y ->
    if not critical.(x).(y) then (
      let boats = liste_bateaux_position (x,y) in
      let my_galleons = List.filter (fun b -> b.deplacable
                                              && b.joueur = !mon_id
                                              && b.btype = Bateau_galion)
                                    (Array.to_list boats) in
      if my_galleons <> [] then
        let target =
          let min_by_dist (k,l) (i,j) =
            if dist_to_critical.(i).(j) < dist_to_critical.(k).(l)
            then (i,j) else (k,l) in
          foldl1 min_by_dist
                 (shuffle_list (List.filter (fun (i,j) -> not critical.(i).(j))
                                            (norm_1_ball (x,y)))) in
        Printf.printf "goto %d %d\n" (fst target) (snd target);
        move_all my_galleons target
    ));


  (*** amasser armée ***)
  let rec recruit_galleons ile =
    if info_ile_or ile >= galion_cout && info_terrain ile = Terrain_ile then (
      match construire Bateau_galion ile with
      | Limite_atteinte -> ()
      | err -> ( log_error "4" err; recruit_galleons ile )
    )
  in
  Array.iter recruit_galleons (mes_iles ());

  flush_outputs ()

(*
** Fonction appelée à la fin de la partie
*)
let partie_fin () =

  flush_outputs ()



(****************************************************)

;;

(* /!\ Ne touche pas a ce qui suit /!\ *)
Callback.register "ml_partie_init" partie_init;;
Callback.register "ml_jouer_tour" jouer_tour;;
Callback.register "ml_partie_fin" partie_fin;;
